"""
This package contains classes for representation of HDL
languages (hdl types, RtlSignal, HdlAssignmentContainer, HdlStmCodeBlockContainer etc.).

It also contains classes for representation of complex HW structures
like dense transaction template (TransTmpl, FrameTmpl).
"""
# [TODO] derive from hdlConvertorAst.hdlAst